#ifndef ITER_H_SEEN
#define ITER_H_SEEN


// data structures: streams and iteratees

typedef struct Stream_ {
    enum {CHUNK, END} state;

    union {
        // CHUNK
        struct {
            void *elems;        // array of elements
            size_t elemsize;    // size in bytes of each element
            size_t n;           // number of elements
            void *orig;         // original array if 'elems' has been advanced
        };

        // END
        const char *error;  // NULL = normal end
    };
} Stream;

// NB: Iteratee is used as a value type.
typedef struct Iteratee_ Iteratee;
struct Iteratee_ {
    enum {DONE, CONT, STOP} state;
    
    union {
        // DONE
        void *result;

        // CONT
        struct {
            Iteratee (*cont)(void *env, Stream *input);
            void *env;

            const char *error;  // STOP only
        };
    };
};


// running iteratees

void *finish(Iteratee iter);
void *run(Iteratee iter);       // an alias for 'finish'

// building iteratees

Iteratee done(void *result);
Iteratee cont(Iteratee (*f)(void *, Stream *), void *env);
Iteratee stop(const char *error, Iteratee (*resume)(void *, Stream *), void *env);
Iteratee fail(const char *error);

Iteratee bind(Iteratee iter, Iteratee (*k)(void *env, void *result), void *env);
Iteratee bind_(Iteratee i, Iteratee j);
Iteratee seq(size_t n, ...);
Iteratee seq_(size_t n, ...);

// primitive iteratees

extern const Iteratee head;
extern const Iteratee chunk;
extern const Iteratee count;
extern const Iteratee drop;
extern const Iteratee take;
extern const Iteratee take1;
extern const Iteratee drop1;


// enumerators

typedef struct Enumerator_ {
    Iteratee (*run)(void *env, Iteratee iter);
    void *env;
} Enumerator;

Enumerator enumerator(Iteratee (*run)(void *, Iteratee), void *env);

Iteratee apply(Enumerator en, Iteratee iter);

Enumerator append(Enumerator a, Enumerator b);

Enumerator enumf(FILE *f);
Enumerator enumfile(const char *path);


// enumeratees

typedef Enumerator Enumeratee;  // let user convey the distinction, even
                                // if the type system can't make it.

Enumeratee decode(Iteratee elemiter);
Enumeratee filter(int (*pred)(void *elem));
Enumeratee ewhile(int (*pred)(void *elem));
Enumeratee until(int (*pred)(void *elem));

// takes an iteratee that returns an iteratee
// fuse the inner iteratee into it, i.e.:
// finish() the inner iteratee when the outer finishes
// returns an iteratee that yields the result of the inner one
Iteratee fuse(Iteratee ie);

// wrap an iteratee in an enumeratee, i.e:
// apply ee to ie and fuse the nested iteratees
Iteratee wrap(Enumeratee ee, Iteratee ie);


#endif // ITER_H_SEEN
